const PACKAGES = [
	'd3-array',
	'd3-axis',
	'd3-brush',
	'd3-chord',
	'd3-color',
	'd3-contour',
	'd3-delaunay',
	'd3-dispatch',
	'd3-drag',
	'd3-dsv',
	'd3-ease',
	'd3-fetch',
	'd3-force',
	'd3-format',
	'd3-geo',
	'd3-hierarchy',
	'd3-interpolate',
	'd3-path',
	'd3-polygon',
	'd3-quadtree',
	'd3-random',
	'd3-scale',
	'd3-scale-chromatic',
	'd3-selection',
	'd3-shape',
	'd3-time',
	'd3-time-format',
	'd3-timer',
	'd3-transition',
	'd3-zoom',
];

function find(importDeclarationPath) {
	const path = importDeclarationPath.node.source.value;
	const found = PACKAGES.find(p => path === p);
	return found;
}

function getLastNameInPath(path) {
	const splited = path.split('/');
	if (splited.length === 1) {
		return false;
	}
	return splited.pop();
}

function sortImports(a, b) {
	if (a.type === 'ImportDefaultSpecifier') {
		return -1;
	}
	if (b.type === 'ImportDefaultSpecifier') {
		return 1;
	}
	return 0;
}

module.exports = function transform({ types }) {
	return {
		visitor: {
			Program(path) {
				const importSpecifiers = {};
				const requireCalls = [];
				let lastImport;
				const addConst = {};
				let realLastImport;

				path.traverse({
					CallExpression: {
						exit(callExpression) {
							if (callExpression.node.callee.name === 'require') {
								if (PACKAGES.indexOf(callExpression.node.arguments[0].value) !== -1) {
									const mod = callExpression.node.arguments[0].value.replace('d3-', '');
									callExpression.node.arguments[0] = types.stringLiteral('d3');
									// we must wrap callexpression into memberexpression
									requireCalls.push({
										callExpression,
										replace: types.memberExpression(
											callExpression.node,
											types.identifier(mod),
										)
									});
									callExpression.remove();
								}
							}
						}
					},
					ImportDeclaration: {
						exit(importDeclarationPath) {
							realLastImport = importDeclarationPath;
							const packageName = find(importDeclarationPath);
							if (packageName && !importSpecifiers['d3']) {
								importSpecifiers['d3'] = [];
							}
							if (packageName) {
								importSpecifiers['d3'].push(
									...importDeclarationPath.node.specifiers.map(({ local, imported, type }) => {
										const localName = local.name;
										let importedName = localName;
										if (imported) {
											importedName = imported.name;
										}
										return types.importSpecifier(
											types.identifier(localName),
											types.identifier(importedName),
										);
									}),
								);
								importDeclarationPath.remove();
							}
							lastImport = importDeclarationPath;
						},
					},
				});

				requireCalls.forEach(info => {
					info.callExpression.insertAfter(info.replace);
				});

				const keys = Object.keys(importSpecifiers);
				if (keys.length > 0) {
					keys.forEach(packageName => {
						const source = types.stringLiteral('d3');
						const imp = types.importDeclaration(
							importSpecifiers[packageName].filter(Boolean).sort(sortImports),
							source,
						);

						const extraKeys = Object.keys(addConst[packageName] || {});
						if (extraKeys.length > 0) {
							extraKeys.forEach(extraKey => {
								const properties = addConst[packageName][extraKey].map(n =>
									types.objectProperty(types.identifier(n), types.identifier(n), false, true),
								);
								const pattern = types.objectPattern(properties);
								const declarator = types.variableDeclarator(
									pattern,
									types.identifier(`__${extraKey}`),
								);
								const c = types.variableDeclaration('const', [declarator]);
								realLastImport.insertAfter(c);
							});
						}
						if (lastImport) {
							lastImport.insertAfter(imp);
						}
					});
				}
			},
		},
	};
};
